from typing import Any, Callable, Dict
import json

from arrow import get as _


def value_as_callable(value) -> Callable:
    def func():
        return value

    return func


class MemoryModel:
    def __init__(self, **raw_data):
        self.__data__ = raw_data
        self.load(**raw_data)

    def load(self, **raw_data):
        for m, value in raw_data.items():
            setattr(self, f"resolve_{m}", value_as_callable(value))

    def get_resolvers(self) -> Dict:
        _resolvers = {}
        for method in dir(self):
            if method.startswith("resolve_"):
                field_name = method.replace("resolve_", "")
                _resolvers.update({field_name: getattr(self, method)})
        return _resolvers

    def get_field_resolver(self, field_name) -> Callable:
        resolvers = self.get_resolvers()
        try:
            return resolvers[field_name]
        except KeyError:
            raise AttributeError(f'has not method "{field_name}"')

    def __getattr__(self, item) -> Any:
        field_resolver = self.get_field_resolver(item)
        return field_resolver()

    def __str__(self) -> str:
        # TODO: resolve -> Complexity too high!
        fields = {
            name: func() if not isinstance(func, type(self)) else f"{func()}"
            for name, func in self.get_resolvers().items()
        }
        return json.dumps(fields)


class Profile(MemoryModel):
    """User profile model"""


class User(MemoryModel):
    """User memory model"""

    scopes: list = None
    token: str = None
    is_anonymous: bool = True
    has_all_api_scopes: bool = None
    token_expire_at: str = None
    token_created_at: str = None
    profile: dict = None

    def resolve_has_all_api_scopes(self) -> bool:
        required_als = getattr(self, "api_level_required_scopes", [])
        if not required_als:
            return True
        scopes = self.resolve_scopes()
        return required_als in scopes

    def resolve_token_expire_at(self) -> (str, None):
        """Parse token expiration date"""
        claims = self.resolve_claims()
        if claims:
            expiration_date = claims.get("exp")
            if expiration_date:
                return _(expiration_date).format("YYYY-MM-DD HH:mm:ss")
        return self.token_expire_at

    def resolve_token_created_at(self) -> (str, None):
        """Parse token creation date"""
        claims = self.resolve_claims()
        if claims:
            expiration_date = claims.get("iat")
            if expiration_date:
                return _(expiration_date).format("YYYY-MM-DD HH:mm:ss")
        return None

    def resolve_scopes(self) -> list:
        """Get token scopes"""
        claims = getattr(self, "claims", {})
        scopes = claims.get("scope", "")
        return scopes.split(",") if scopes else []

    def resolve_is_anonymous(self) -> list:
        """Is anonymous session"""
        return self.resolve_token() is None

    def resolve_profile(self) -> Profile:
        if self.profile:
            return Profile(**self.profile)



from graphene import ObjectType, String, Schema

from tokko_auth.decorators import has_permission


class QueryTesting(ObjectType):
    protected_query = String()
    protected_query_two_required_scopes = String()
    protected_query_at_least_one_of_required_scopes = String()

    @has_permission("test:testing", has_all=True)
    def resolve_protected_query(self, info):
        return "You're Authorized!"

    @has_permission("test:testing", "my_other:scope", has_all=True)
    def resolve_protected_query_two_required_scopes(self, info):
        return "You're Authorized!"

    @has_permission("test:testing", "my_other:scope", at_least_one=True)
    def resolve_protected_query_at_least_one_of_required_scopes(self, info):
        return "You're Authorized!"


TESTING_SCHEMA = Schema(query=QueryTesting, auto_camelcase=False)



from Crypto.PublicKey import RSA
from arrow import get as _, now
from jose import jwt


RSA_KEY = RSA.generate(bits=1024)
RSA_ID = RSA_KEY.export_key().decode()
RSA_PUB = RSA_KEY.publickey().publickey().export_key().decode()

PAYLOAD = {
    "scope": "test:testing",
    "permissions": ["test:testing"],
    "iss": "https://testing.provider.com/",
    "aud": "https://testing.com/service",
    "iat": (now() - _("1970-01-01")).total_seconds(),
    "exp": (now().shift(days=1) - _("1970-01-01")).total_seconds(),
    "gty": "client-credentials",
}

WITHOUT_SCOPES_PAYLOAD = {
    "iss": "https://testing.provider.com/",
    "aud": "https://testing.com/service",
    "iat": (now() - _("1970-01-01")).total_seconds(),
    "exp": (now().shift(days=1) - _("1970-01-01")).total_seconds(),
    "gty": "client-credentials",
}

EMPTY_PAYLOAD = {}


ALGORITHM = "RS256"

TOKEN_HEADERS = {"kid": "my-k-id"}


VALID_TOKEN = jwt.encode(
    PAYLOAD, key=RSA_ID, algorithm=ALGORITHM, headers=TOKEN_HEADERS
)

INVALID_TOKEN_NO_SCOPES = jwt.encode(
    WITHOUT_SCOPES_PAYLOAD, key=RSA_ID, algorithm=ALGORITHM, headers=TOKEN_HEADERS
)

INVALID_TOKEN_NO_CLAIMS = jwt.encode(
    EMPTY_PAYLOAD, key=RSA_ID, algorithm=ALGORITHM, headers=TOKEN_HEADERS
)



import logging

from tokko_auth.constants import UNAUTHORIZED, INTERNAL_ERROR, GATEWAY_ERROR

logger = logging.getLogger(__name__)


class Error:
    """Error representation"""

    def __init__(self, code: str, description: str):
        self.code = code
        self.description = description

    def __str__(self) -> str:
        return f'"{self.code.upper()}". {self.description}'


class HeadersError(Error):
    def __init__(self, description: str):
        super().__init__("header_error", description)


class SettingsError(Error):
    def __init__(self, description: str):
        super().__init__("settings_error", description)


class ServiceError(Error):
    def __init__(self, description: str):
        super().__init__("service_error", description)


class TokenError(Error):
    def __init__(self, description: str):
        super().__init__("token_error", description)


class UserError(Error):
    def __init__(self, description: str):
        super().__init__("user_error", description)


class ScopesError(Error):
    def __init__(self, description: str):
        super().__init__("scopes_error", description)


class ProviderError(Error):
    def __init__(self, description: str):
        super().__init__("provider_error", description)


class AuthPermissionErrors(PermissionError):
    """Authorization Error base"""

    debug_message = ""

    def __init__(self, error: Error, status_code: int):
        self.error = error
        self.status_code = status_code
        debug_msg = f" {self.debug_message}." if self.debug_message else ""
        logger.exception(f"{self}.{debug_msg}")

    @property
    def as_json(self) -> dict:
        return \
            {
                "code": self.error.code,
                "description": self.error.description,
                "status_code": self.status_code
            }

    def __str__(self) -> str:
        return f"{self.status_code} - {self.error}"


class PermissionDenied(AuthPermissionErrors):
    """Permission Denied"""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class NamespaceNotMatch(AuthPermissionErrors):
    """Token not contains required Namespace"""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class ProviderHttpError(AuthPermissionErrors):
    """Authorization unavailable"""

    debug_message = "NON_OK_PROVIDER_RESPONSE"

    def __init__(self, status_code=None, description: str = None):
        error = ProviderError(description or f"{self.__doc__}")
        super().__init__(error, status_code=status_code or GATEWAY_ERROR)


class EnvVarNotInitialized(AuthPermissionErrors):
    """ENV variable not properly initialized"""

    debug_message = "REQUIRED_VARIABLE_NOT_INITIALIZED"

    def __init__(self, var_name: str = None):
        error = ServiceError(
            f"{self.__doc__}" + f". EnvVar: {var_name.upper()}" if var_name else ""
        )
        super().__init__(error, status_code=INTERNAL_ERROR)


class JSONResponseExpected(AuthPermissionErrors):
    """Authorization unavailable"""

    debug_message = "PROVIDER_JSON_RESPONSE_EXPECTED"

    def __init__(self, description: str = None):
        error = ProviderError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class ProviderUnreachable(AuthPermissionErrors):
    """Authorization unavailable"""

    debug_message = "PROVIDER_UNREACHABLE"

    def __init__(self, description: str = None):
        error = ProviderError(description or f"{self.__doc__}")
        super().__init__(error, status_code=INTERNAL_ERROR)


class UserInfoROField(AuthPermissionErrors):
    """Permission Denied"""

    debug_message = "USER_FIELD_IS_READONLY"

    def __init__(self, description: str = None):
        error = ProviderError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class TokenNotFound(AuthPermissionErrors):
    """Authorization JWT not found"""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class MustStartWithBearer(AuthPermissionErrors):
    """Authorization header must start with 'Bearer'"""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class MustBeBearerToken(AuthPermissionErrors):
    """Authorization header must be 'Bearer <auth-token>'"""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class InsufficientScopes(AuthPermissionErrors):
    """Required scope was not found"""

    def __init__(self, description: str = None):
        error = ScopesError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class ScopesNotFound(AuthPermissionErrors):
    """Provided TOKEN not contains SCOPES"""

    def __init__(self, description: str = None):
        error = ScopesError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class ClaimsNotFound(AuthPermissionErrors):
    """Unable to find claims"""

    def __init__(self, description: str = None):
        error = HeadersError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class UnableDecodeToken(AuthPermissionErrors):
    """Unable to decode authentication token"""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class TokenIsExpired(AuthPermissionErrors):
    """Token signature is expired"""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class AuthProviderConnError(AuthPermissionErrors):
    """Authentication provider did not respond"""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class IncorrectClaims(AuthPermissionErrors):
    """Token claims are not correct."""

    def __init__(self, description: str = None):
        error = TokenError(description or f"{self.__doc__}")
        super().__init__(error, status_code=UNAUTHORIZED)


class IncompatibleConfigs(AuthPermissionErrors):
    """Unable to decode authentication token"""

    def __init__(self, description: str = None):
        error = SettingsError(description or f"{self.__doc__}")
        super().__init__(error, status_code=INTERNAL_ERROR)



from dataclasses import dataclass


@dataclass
class AuthHeader:
    token: str = None

    @property
    def authorization(self) -> dict:
        return {'Authorization': f'Bearer {self.token}'}


@dataclass
class Context:
    headers: dict = None



from tokko_auth.exceptions import ScopesNotFound, IncompatibleConfigs, TokenNotFound
from tokko_auth.util import beautiful_errors, auth_is_enable
from tokko_auth.helpers import AuthorizationHelper as Auth
from tokko_auth import settings


def __require_all_scopes__(has_all=None, at_least_one=None) -> bool:
    if all([x is None for x in [has_all, at_least_one]]):
        raise IncompatibleConfigs
    if all([not x for x in [has_all, at_least_one]]):
        raise IncompatibleConfigs
    if at_least_one and has_all:
        return True
    return True if has_all and not at_least_one else False


def has_permission(*scopes, has_all: bool = True, at_least_one: bool = False):
    """
    Decorate Django view functions, methods, GQL Mutates & Queries
    """

    def wrap(func):
        def wrapped_function(*args, **kwargs):
            if func.__name__ == "mutate" or func.__name__.startswith("resolve_"):
                # Use GQL error handler instead
                # auth beautify error tool
                forced = True
                # Resolve GraphQL view request
                request = args[1].context
            else:
                forced = settings.AUTH_BEAUTIFY_ERROR_RESPONSE
                # Resolve DJANGO view request
                request = args[0]

            require_all = any(
                [
                    # Has all scope:
                    has_all and not at_least_one,  # T & F
                    not has_all and not at_least_one,  # F & F
                    # At least one scope:
                    # has_all and at_least_one,  # T & T
                    # not has_all and at_least_one,  # F & T
                    # not has_all and at_least_one,  # F & T
                ]
            )
            try:
                # Check enable/disable rules
                if auth_is_enable:
                    auth = Auth()
                    if not hasattr(request, "headers"):
                        raise TokenNotFound
                    _token = auth.get_token(request.headers)
                    try:
                        auth.has_required_scopes(_token, scopes, require_all)
                    except ScopesNotFound:
                        if not settings.AUTH_FAIL_SAFE_ENABLE:
                            raise ScopesNotFound
                return func(*args, **kwargs)
            except Exception as e:
                return beautiful_errors(e, forced)

        return wrapped_function

    return wrap



from typing import Dict, List

from requests.exceptions import ConnectionError, Timeout, HTTPError
from jose.exceptions import JWTError
from requests import Session
from jose import jwt

# App settings
from tokko_auth import settings


from tokko_auth.exceptions import (
    TokenNotFound,
    UnableDecodeToken,
    ClaimsNotFound,
    ProviderHttpError,
    ProviderUnreachable,
    JSONResponseExpected,
    EnvVarNotInitialized,
    MustBeBearerToken,
    MustStartWithBearer,
    ScopesNotFound,
    InsufficientScopes,
)

# Connections Pool
# I think this not has sense at all ...
# I make just one request ¯\_(ツ)_/¯
pool_conns = Session()

# Types
List_or_Str = (List[str], str)


class AuthorizationHelper:
    """Authorization helper"""

    def namespaces(self, token) -> List[str]:
        scopes = self.scopes(token)
        return [s.split(":")[0] for s in scopes]

    def has_namespace(self, namespace, token) -> bool:
        """Checks: "namespace" is in "token".claims.scopes.namespaces"""
        if not namespace:
            return True
        return namespace in self.namespaces(token)

    def has_required_scopes(self, token: str, required_scopes: List_or_Str, require_all=True) -> bool:
        """Scopes validation"""
        scopes = self.scopes(token)
        if not isinstance(required_scopes, (list, tuple)):
            required_scopes = [required_scopes]
        if scopes:
            required_vs_provided = [scope in scopes for scope in required_scopes]
            if require_all and not all(required_vs_provided):
                raise InsufficientScopes
            elif not require_all and not any(required_vs_provided):
                raise InsufficientScopes
            return True
        else:
            raise ScopesNotFound

    def get_token(self, headers: dict) -> str:
        """Get token from "Authorization" header"""
        auth = self.get_authorization_header(headers)
        parts = auth.split()
        if len(parts) == 1 or len(parts) > 2:
            raise MustStartWithBearer
        # Semantic Token validations
        if not parts[0].lower() == "bearer":
            raise MustBeBearerToken
        return parts[1]

    @staticmethod
    def get_authorization_header(headers: dict) -> str:
        """Get "Authorization" header"""
        if not hasattr(headers, "get"):
            raise TokenNotFound
        auth = headers.get("Authorization") or headers.get("tokko_auth")
        # Syntactic Token validations
        if not auth:
            raise TokenNotFound
        return auth

    def scopes(self, token) -> List[str]:
        """Get Scopes list from claims"""
        claims = self.claims(token)
        scopes = claims.get("permissions")
        if scopes:
            return scopes
        raise ScopesNotFound

    @staticmethod
    def claims(token: str) -> Dict:
        """Get token's unverified claims"""
        try:
            unverified_claims = jwt.get_unverified_claims(token)
        except JWTError:
            raise UnableDecodeToken
        if not unverified_claims:
            raise ClaimsNotFound
        return unverified_claims

    @staticmethod
    def user_properties(token: str) -> Dict:
        """Get user info by token"""
        try:
            auth0_domain = settings.AUTH0_DOMAIN
            if not auth0_domain:
                # Blow up!
                raise EnvVarNotInitialized
            res = pool_conns.get(
                f"https://{auth0_domain}/userinfo",
                headers={"Authorization": f"Bearer {token}"},
            )
            res.raise_for_status()
            content_type = res.headers.get("Content-Type")
            if "application/json" not in content_type:
                raise JSONResponseExpected
            return res.json()
        except (ConnectionError, Timeout):
            raise ProviderUnreachable
        except HTTPError:
            raise ProviderHttpError



from os import environ
from typing import Any
import logging

from django.conf import settings

from tokko_auth.exceptions import EnvVarNotInitialized
from tokko_auth.constants import COGNITO


# Local logger instance
logger = logging.getLogger(__name__)


def env_or_settings(var_name, default=None, raise_error=False) -> Any:
    """Get env var values from ENV or DJ SETTINGS"""
    try:
        _var = environ[var_name]
    except KeyError:
        _var = getattr(settings, var_name, default)
        if _var is None and raise_error:
            exception = EnvVarNotInitialized(var_name)
            logger.exception(f'{exception}')
            raise exception
    return _var


# Auth app settings
AUTH_FAIL_SAFE_ENABLE = env_or_settings("AUTH_FAIL_SAFE_ENABLE", False)
AUTH_USERINFO_ENABLE = env_or_settings("AUTH_USERINFO_ENABLE", False)
AUTH_API_NAMESPACE = env_or_settings("AUTH_API_NAMESPACE", '')
AUTH0_DOMAIN = env_or_settings("AUTH0_DOMAIN", raise_error=True)
AUTH_BEAUTIFY_ERROR_RESPONSE = env_or_settings("AUTH_BEAUTIFY_ERROR_RESPONSE", True)
AUTH_FULL_DISABLED = env_or_settings("AUTH_FULL_DISABLED", False)
AUTH_ALLOW_PRODUCTIVE_SHUTDOWN = env_or_settings(
    "AUTH_ALLOW_PRODUCTIVE_SHUTDOWN", False
)
SAMPLES_ARE_ENABLE = env_or_settings("SAMPLES_ARE_ENABLE", True)


ALGORITHMS = environ.get("TOKKO_AUTH_ALGORITHMS",
                         # Pre deprecation support
                         environ.get("ALGORITHMS"))

API_AUDIENCE = environ.get("TOKKO_AUTH_API_AUDIENCE",
                           # Pre deprecation support
                           environ.get("API_AUDIENCE"))

AUTH_DOMAIN = environ.get("TOKKO_AUTH_PROVIDER_DOMAIN",
                          # Pre deprecation support
                          environ.get("AUTH0_DOMAIN"))

AUTH_STRATEGY = environ["TOKKO_AUTH_STRATEGY"] = COGNITO



import os
from setuptools import find_packages, setup

readme_file = os.path.join(os.path.dirname(__file__), 'README.md')

with open(readme_file) as readme:
    README = f"{readme.read()}"
README = README.replace(README.split('# Install')[0], '')

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setup(
    name="tokko_auth",
    version="0.0.7.4",
    packages=find_packages(),
    include_package_data=True,
    license="BSD License",
    description="Tokko Auth2 flavor.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/TokkoLabs/authorization.git",
    author="Jose Salgado",
    author_email="jsalgado@navent.com",
    install_requires=[
        "arrow",
        "python-jose",
        "requests",
        "Django",
        "crypto",
        "pycryptodome",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



import unittest

from django.test import TestCase
from graphene.test import Client

from tokko_auth.settings import (
    SAMPLES_ARE_ENABLE as using_samples,
    AUTH_API_NAMESPACE as api_namespace,
    AUTH_FULL_DISABLED as is_disabled
)
from tokko_auth.exceptions import TokenNotFound, ScopesNotFound, InsufficientScopes

from .schema import TESTING_SCHEMA
from .crypto import VALID_TOKEN, INVALID_TOKEN_NO_SCOPES
from .context import Context, AuthHeader


class ServiceTest(TestCase):
    """Service Tests"""

    def setUp(self) -> None:
        self.valid_token = VALID_TOKEN
        self.invalid_token_no_scopes = INVALID_TOKEN_NO_SCOPES
        self.gql_client = Client(TESTING_SCHEMA)

    def test_01_request_protected_query(self):
        ctx = Context(headers=AuthHeader(token=self.valid_token).authorization)
        res = self.gql_client.execute("query {protected_query}", context=ctx)
        self.assertEquals(res, {"data": {"protected_query": "You're Authorized!"}})

    def test_02_request_protected_query_without_credentials(self):
        res = self.gql_client.execute("query {protected_query}")
        err_message = res["errors"][0]["message"]
        self.assertEquals(err_message, f"{TokenNotFound()}")
        self.assertEquals(res["data"]["protected_query"], None)

    def test_03_request_protected_query_without_scopes(self):
        ctx = Context(
            headers=AuthHeader(token=self.invalid_token_no_scopes).authorization
        )
        res = self.gql_client.execute("query {protected_query}", context=ctx)
        err_message = res["errors"][0]["message"]
        self.assertEquals(err_message, f"{ScopesNotFound()}")
        self.assertEquals(res["data"]["protected_query"], None)

    def test_04_request_protected_query_without_all_required_scopes(self):
        ctx = Context(headers=AuthHeader(token=self.valid_token).authorization)
        res = self.gql_client.execute(
            "query {protected_query_two_required_scopes}", context=ctx
        )
        err_message = res["errors"][0]["message"]
        self.assertEquals(err_message, f"{InsufficientScopes()}")
        self.assertEquals(res["data"]["protected_query_two_required_scopes"], None)

    def test_05_request_protected_query_one_of_required_scopes(self):
        ctx = Context(headers=AuthHeader(token=self.valid_token).authorization)
        res = self.gql_client.execute(
            "query {protected_query_at_least_one_of_required_scopes}", context=ctx
        )
        self.assertEquals(
            res,
            {
                "data": {
                    "protected_query_at_least_one_of_required_scopes": "You're Authorized!"
                }
            },
        )

    @unittest.skipIf(not using_samples, "Samples are disabled")
    def test_06_request_status_data(self):
        r = self.client.get("/auth/status/")
        _is_enable = r.json()["settings"]["is_enable"]
        _api_namespace = r.json()["settings"]["api_namespace"]
        self.assertEquals(_is_enable, not is_disabled)
        self.assertEquals(_api_namespace, api_namespace)

    @unittest.skipIf(not using_samples, "Samples are disabled")
    def test_07_request_protected_endpoint(self):
        r = self.client.get("/auth/protected/", HTTP_AUTHORIZATION=f'Bearer {self.valid_token}')
        self.assertEquals(r.status_code, 200)
        self.assertEquals(r.json(), {"message": "Authorized!"})
